import 'dart:async';

/// The first arguments is a `List<dynamic>` of the positioned arguments,
/// and the second arguments is a `Map<Symbol, dynamic>` of the named arguments.
/// If you want to access a named argument, you should use the `Symbol`
/// instead of the variable name which is usually a `String`.
/// See the below examples in the `example/` directory.
///
/// ```dart
/// int echo(List positioned, Map named) {
///  if (named.containsKey('key')) {
///     // This condition does not pass the test.
///     // You can not use a `String` to a key.
///     print(named['key']);
///   }
///   // The first method to access the named arguments.
///   if (named.containsKey(#key2)) {
///     print(named[#key2]);
///   }
///   // The second method to access the named arguments.
///   if (named.containsKey(Symbol('key3'))) {
///     print(named[#key3]);
///   }
///   return positioned.length + named.length;
/// }
/// ```
typedef VarArgsHandler<T> = FutureOr<T> Function(List, Map);

class VarArgsFunction<T> {
  VarArgsFunction(this.onCall);

  final VarArgsHandler<T> onCall;

  @override
  FutureOr<T> noSuchMethod(Invocation info) {
    if (!info.isAccessor && info.memberName == Symbol('call')) {
      return onCall?.call(info.positionalArguments, info.namedArguments);
    } else {
      throw NoSuchMethodError.withInvocation(this, info);
    }
  }
}

extension VarArgsTransformer<T> on VarArgsHandler {
  /// Intentionally use `dynamic` so that dart analyzer and
  /// dart compiler can not catch the unimplemented method call.
  dynamic get varArgs => VarArgsFunction(this);
}
